// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Fischer
#include <cstdlib>		 // for exit, EXIT_FAILURE, EXIT_SUC...
#include <functional>	 // bind
#include <iostream>		 // for operator<<, basic_ostream
#include <thread>			 // for hardware_concurrency

#include "Actor.hpp"										 // for Actor
#include "BinaryPrivateSemaphore.hpp"		 // for BPS
#include "CountingPrivateSemaphore.hpp"	 // for CPS
#include "Fiber.hpp"										 // for Fiber
#include "Runtime.hpp"									 // for Runtime
#include "emper.hpp"										 // for spawn

class AlarmActor : public Actor<BPS*> {
 protected:
	void receive(BPS* sem) override { sem->signal(); }

 public:
	AlarmActor(Runtime& runtime) : Actor(runtime) {}
	void stop() { Actor::stop(); }
};

auto main(int argc, char* argv[]) -> int {
	unsigned int sleeper_count = std::thread::hardware_concurrency();
	unsigned int sleeps = 1000;

	if (argc > 3) {
		std::cerr << "Usage: " << argv[0] << " [fiber count] [block count]" << std::endl;
		exit(EXIT_FAILURE);
	}

	const int DECIMAL = 10;
	if (argc > 1) {
		sleeper_count = strtol(argv[1], nullptr, DECIMAL);
	}

	if (argc > 2) {
		sleeps = strtol(argv[2], nullptr, DECIMAL);
	}

	Runtime runtime;

	Fiber* fiber = Fiber::from([&] {
		AlarmActor alarmActor(runtime);
		alarmActor.start();

		CPS cps;
		for (unsigned int i = 0; i < sleeper_count; ++i) {
			spawn(
					[&alarmActor, &sleeps] {
						for (unsigned int i = 1; i <= sleeps; ++i) {
							BPS sem;
							alarmActor.tell(&sem);
							sem.wait();
						}
					},
					cps);
		}

		// Wait for the sleeping fibers to finish
		cps.wait();

		// Wait for the actor to become idle.
		bool actorIdle = alarmActor.waitUntilIdle(60 * 1000);
		if (!actorIdle) {
			std::cerr << "FAILURE: Actor did not went idle";
			exit(EXIT_FAILURE);
		}

		alarmActor.stop();

		exit(EXIT_SUCCESS);
	});

	runtime.scheduleFromAnywhere(*fiber);

	runtime.waitUntilFinished();

	return EXIT_FAILURE;
}
